# Dependencies
from shapely.geometry import mapping
import fiona  # type: ignore

# Relative Imports
from .custom_types import Path, PathLike
from .spec1D_models import GeoSpectrum1D
from .specgroup_models import SpectrumGroup
from .reading import read_geodata


def make_points(
    spectra: list[GeoSpectrum1D] | SpectrumGroup,
    output_path: PathLike,
    group_geodata_fp: PathLike | None = None,
):
    schema = {
        "geometry": "Point",
        "properties": {"name": "str", "id": "int"},
    }
    fiona_config = {"crs": None, "driver": "ESRI Shapefile", "schema": schema}
    if isinstance(spectra, SpectrumGroup):
        if group_geodata_fp is None:
            raise ValueError(
                "If a Spectrum Group is used to make points, a geodata path"
                " must be provided."
            )
        spectra_list: list[GeoSpectrum1D] = []
        geodata = read_geodata(geodata_fp=group_geodata_fp)
        for pnt_spec in spectra.spectra:
            geospec = GeoSpectrum1D.from_point_spec(geodata, pnt_spec)
            spectra_list.append(geospec)
    elif isinstance(spectra, list):
        spectra_list = spectra

    if Path(output_path).is_dir():
        for spec in spectra_list:
            fiona_config["crs"] = spec.point.crs
            save_file = Path(output_path, spec.name).with_suffix(".shp")
            with fiona.open(save_file, "w", **fiona_config) as c:
                c.write(
                    {
                        "geometry": mapping(spec.shapely_geometry()),
                        "properties": {"name": spec.name, "id": 1},
                    }
                )
    elif Path(output_path):
        save_file = Path(output_path).with_suffix(".shp")
        fiona_config["crs"] = spectra_list[0].point.crs
        with fiona.open(save_file, "w", **fiona_config) as c:
            for n, spec in enumerate(spectra_list):
                c.write(
                    {
                        "geometry": mapping(spec.shapely_geometry()),
                        "properties": {"name": spec.name, "id": n},
                    }
                )


def make_polygons(
    spectra: list[SpectrumGroup], geodata_fp: PathLike, output_path: PathLike
):
    geodata = read_geodata(geodata_fp)
    schema = {
        "geometry": "Polygon",
        "properties": {"name": "str", "id": "int", "nspectra": "int"},
    }
    fiona_config = {
        "crs": geodata.crs,
        "driver": "ESRI Shapefile",
        "schema": schema,
    }
    if Path(output_path).is_dir():
        for spec in spectra:
            save_file = Path(output_path, spec.name).with_suffix(".shp")
            with fiona.open(save_file, "w", **fiona_config) as c:
                c.write(
                    {
                        "geometry": mapping(spec.shapely_geometry(geodata)),
                        "properties": {
                            "name": spec.name,
                            "id": 1,
                            "nspectra": spec.nspectra,
                        },
                    }
                )
    elif Path(output_path):
        save_file = Path(output_path).with_suffix(".shp")
        with fiona.open(save_file, "w", **fiona_config) as c:
            for n, spec in enumerate(spectra):
                c.write(
                    {
                        "geometry": mapping(spec.shapely_geometry(geodata)),
                        "properties": {
                            "name": spec.name,
                            "id": n,
                            "nspectra": spec.nspectra,
                        },
                    }
                )



# Standard Libraries
from enum import StrEnum
from typing import Literal

# Dependencies
from pydantic import BaseModel, Field
import numpy as np
from shapely.geometry import Point

# Relative Imports
from .geospatial_models import (
    BaseGeolocationModel,
    PointGeolocation,
    PointModel,
)
from .wvl_models import WvlModel

type Spec1DFileLiteral = Literal["rawspec"] | Literal["pntspec"] | Literal[
    "geospec"
]


class Spec1DFileTypes(StrEnum):
    RAW = ".rawspec"
    PNT = ".pntspec"
    GEO = ".geospec"


class Spectrum1D(BaseModel):
    """
    Object representing a single spectrum.

    Attributes
    ----------
    name: str
        Name of the spectrum.
    spectrum: list[float]
        Data of the spectrum (e.g. Reflectance, Emissivity, Transmittance)
    wavelength: WvlModel
        Wavelength Object corresponding to the spectrum.

    Notes
    -----
    See wvl_models.py for information on the Wavelength Object.
    """

    name: str
    spectrum: list[float]
    wavelength: WvlModel
    bbl_applied: bool = Field(default=False)

    @classmethod
    def empty(cls) -> "Spectrum1D":
        return cls(
            name="null",
            spectrum=[],
            wavelength=WvlModel.default_bbl([0], "um"),
        )

    def applybbl(self):
        self.spectrum = list(np.asarray(self.spectrum)[self.wavelength.bbl])
        self.bbl_applied = True

    def tonumpy(self):
        return np.asarray(self.spectrum, dtype=np.float32)


class PointSpectrum1D(Spectrum1D):
    """
    Object representing a spectrum pulled from a non-georeferenced spectral
    cube.

    Attributes
    ----------
    name: str
        Name of the spectrum.
    spectrum: list[float]
        Data of the spectrum (e.g. Reflectance, Emissivity, Transmittance)
    wavelength: WvlModel
        Wavelength Object corresponding to the spectrum.
    point: PointModel
        Just a standard X, Y ordered pair object.

    Notes
    -----
    See wvl_models.py for information on the Wavelength Object.
    See geospatial_models.py for information on the PointModel Object.
    """

    pixel: PointModel

    @classmethod
    def from_pixel_coord(
        cls, x: float, y: float, spec1d: Spectrum1D
    ) -> "PointSpectrum1D":
        return cls(
            name=spec1d.name,
            spectrum=spec1d.spectrum,
            wavelength=spec1d.wavelength,
            pixel=PointModel(x=x, y=y),
        )


class GeoSpectrum1D(Spectrum1D):
    """
    Object representing a spectrum pulled from a non-georeferenced spectral
    cube.

    Attributes
    ----------
    name: str
        Name of the spectrum.
    spectrum: list[float]
        Data of the spectrum (e.g. Reflectance, Emissivity, Transmittance)
    wavelength: WvlModel
        Wavelength Object corresponding to the spectrum.
    point: PointGeolocation
        Just a standard X, Y ordered pair object.

    Notes
    -----
    See wvl_models.py for information on the Wavelength Object.
    See geospatial_models.py for information on the PointModel Object.
    """

    point: PointGeolocation

    @classmethod
    def from_pixel_coord(
        cls,
        x: float,
        y: float,
        geoloc: BaseGeolocationModel,
        spec1d: Spectrum1D,
    ):
        return cls(
            name=spec1d.name,
            spectrum=spec1d.spectrum,
            wavelength=spec1d.wavelength,
            point=PointGeolocation.from_base(
                geoloc=geoloc, location=(x, y), location_type="pixel"
            ),
        )

    @classmethod
    def from_map_coord(
        cls,
        x: float,
        y: float,
        geoloc: BaseGeolocationModel,
        spec1d: Spectrum1D,
    ) -> "GeoSpectrum1D":
        return cls(
            name=spec1d.name,
            spectrum=spec1d.spectrum,
            wavelength=spec1d.wavelength,
            point=PointGeolocation.from_base(
                geoloc=geoloc, location=(x, y), location_type="map"
            ),
        )

    @classmethod
    def from_point_spec(
        cls, geoloc: BaseGeolocationModel, spec1d: PointSpectrum1D
    ) -> "GeoSpectrum1D":
        return cls(
            name=spec1d.name,
            spectrum=spec1d.spectrum,
            wavelength=spec1d.wavelength,
            point=PointGeolocation.from_base(
                geoloc=geoloc,
                location=(spec1d.pixel.x, spec1d.pixel.y),
                location_type="pixel",
            ),
        )

    def location_str(self) -> str:
        """Returns a formatted string of location data."""
        return (
            f"({self.point.pixel_point.y}, {self.point.pixel_point.x})  --> "
            f"({self.point.map_point.y:.2f}, {self.point.map_point.y:.2f})"
        )

    def map_location(self) -> tuple[float, float]:
        """Returns a tuple containing the map point (y, x)"""
        return (self.point.map_point.y, self.point.map_point.x)

    def pixel_location(self) -> tuple[float, float]:
        """Returns a tuple containing the pixel point (x, y)"""
        return (self.point.pixel_point.x, self.point.pixel_point.y)

    def shapely_geometry(self) -> Point:
        return Point(self.point.map_point.astuple())



# Dependencies
from dataclasses import dataclass, field
from pydantic import BaseModel, Field, model_validator
import numpy as np
from shapely.geometry import Polygon
from alphashape import alphashape  # type: ignore

# Relative Imports
from .custom_types import PathLike, Path
from .spec1D_models import PointSpectrum1D, Spec1DFileTypes
from .geospatial_models import BaseGeolocationModel
from .wvl_models import WvlModel

# from .geospatial_models import Bounds


@dataclass
class GroupStats:
    mean: np.ndarray
    median: np.ndarray
    stdev: np.ndarray
    error_bounds: tuple[np.ndarray, np.ndarray] = field(init=False)

    def __post_init__(self):
        self.error_bounds = (self.mean - self.stdev, self.mean + self.stdev)


class SpectrumGroup(BaseModel):
    name: str
    spectra: list[PointSpectrum1D]
    spectra_pts: list[tuple[int, int]]
    wavelength: WvlModel
    polygon_vertices: list[tuple[float, float]] = Field(default=[])
    nspectra: int = Field(default=0)
    bbl_applied: bool = Field(default=False)

    @model_validator(mode="after")
    def set_nspectra(self) -> "SpectrumGroup":
        if self.name == "null":
            return self
        self.nspectra = len(self.spectra)
        poly: Polygon = alphashape(self.spectra_pts, alpha=0.9)  # type: ignore
        self.polygon_vertices = [
            (i, j) for i, j in zip(poly.exterior.xy[0], poly.exterior.xy[1])
        ]
        return self

    @classmethod
    def empty(cls) -> "SpectrumGroup":
        return cls(
            name="null",
            spectra=[],
            spectra_pts=[],
            wavelength=WvlModel.default_bbl([0], "um"),
        )

    def resolve_polygon(self) -> None:
        if self.name == "null":
            raise UserWarning(
                "SpectrumGroup was not resolved because its name was not set."
            )
        self.nspectra = len(self.spectra)
        poly: Polygon = alphashape(self.spectra_pts, alpha=0.9)  # type: ignore
        self.polygon_vertices = [
            (i, j) for i, j in zip(poly.exterior.xy[0], poly.exterior.xy[1])
        ]

    def applybbl(self):
        for i in self.spectra:
            i.applybbl()
        self.bbl_applied = True

    def asarray(self) -> np.ndarray:
        if self.bbl_applied:
            _nbands = self.spectra[0].wavelength.ngoodbands
        else:
            _nbands = self.spectra[0].wavelength.nbands
        group_array = np.empty((self.nspectra, _nbands), dtype=np.float32)
        for n, i in enumerate(self.spectra):
            group_array[n, :] = i.spectrum
        return group_array

    def export_to_directory(self, out_dir: PathLike) -> None:
        for i in self.spectra:
            json_str = i.model_dump_json(indent=2)
            with open(
                Path(out_dir, i.name).with_suffix(Spec1DFileTypes.PNT)
            ) as f:
                f.write(json_str)

    def create_mask(self, height: int, width: int) -> np.ndarray:
        mask_arr = np.zeros((height, width), dtype=np.bool)
        x_pts: list[int] = []
        y_pts: list[int] = []
        for spec in self.spectra:
            x_pts.append(int(spec.pixel.x))
            y_pts.append(int(spec.pixel.y))
        x_pts_arr = np.asarray(x_pts)
        y_pts_arr = np.asarray(y_pts)

        mask_arr[x_pts_arr, y_pts_arr] = True

        return mask_arr

    def get_stats(self) -> GroupStats:
        arr = self.asarray()

        return GroupStats(
            mean=np.mean(arr, axis=0),
            median=np.median(arr, axis=0),
            stdev=np.std(arr, axis=0, ddof=1),
        )

    def get_vertices_arr(self) -> np.ndarray:
        vert_arr = np.empty((len(self.polygon_vertices), 2))

        for n, i in enumerate(self.polygon_vertices):
            vert_arr[n, :] = i

        return vert_arr

    def shapely_geometry(self, geodata: BaseGeolocationModel) -> Polygon:
        new_verts: list[tuple[float, float]] = []
        for vert in self.polygon_vertices:
            new_verts.append(geodata.geotransform.pixel_to_map(*vert))
        return Polygon(new_verts)



# Standard Libraries
import logging
from typing import overload, Literal

# Relative Imports
from .wvl_models import WvlModel
from .spec1D_models import (
    Spectrum1D,
    GeoSpectrum1D,
    PointSpectrum1D,
    Spec1DFileLiteral,
)
from .spec3D_models import Spectrum3D, GeoSpectrum3D, Spec3DFileLiteral
from .specgroup_models import SpectrumGroup
from .geospatial_models import BaseGeolocationModel
from ._errors import FileTypeError
from .custom_types import PathLike, Path

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def read_geodata(geodata_fp: PathLike) -> BaseGeolocationModel:
    file_suff = Path(geodata_fp).suffix
    if file_suff != ".geodata":
        raise FileTypeError(
            f"The file type should be .geodata not {file_suff}."
        )
    with open(geodata_fp, "r") as f:
        json_str = f.read()
    return BaseGeolocationModel.model_validate_json(json_str)


@overload
def read_spec1D(
    spec1d_fp: PathLike, kind: Literal["rawspec"]
) -> Spectrum1D: ...


@overload
def read_spec1D(
    spec1d_fp: PathLike, kind: Literal["pntspec"]
) -> PointSpectrum1D: ...


@overload
def read_spec1D(
    spec1d_fp: PathLike, kind: Literal["geospec"]
) -> GeoSpectrum1D: ...


def read_spec1D(spec1d_fp: PathLike, kind: Spec1DFileLiteral):
    if kind == "rawspec":
        with open(spec1d_fp, "r") as f:
            json_str = f.read()
        spec_model = Spectrum1D.model_validate_json(json_str)
        return spec_model
    elif kind == "pntspec":
        with open(spec1d_fp, "r") as f:
            json_str = f.read()
        spec_model = PointSpectrum1D.model_validate_json(json_str)
        return spec_model
    elif kind == "geospec":
        with open(spec1d_fp, "r") as f:
            json_str = f.read()
        spec_model = GeoSpectrum1D.model_validate_json(json_str)
        return spec_model


def read_group(specgroup_fp: PathLike):
    file_suff = Path(specgroup_fp).suffix
    if file_suff != ".specgrp":
        raise FileTypeError(f"The file type should be .wvl not {file_suff}.")
    with open(specgroup_fp, "r") as f:
        json_str = f.read()
    group_model = SpectrumGroup.model_validate_json(json_str)
    return group_model


@overload
def read_spec3D(spec3d_fp: PathLike, kind: Literal["spcub"]) -> Spectrum3D: ...


@overload
def read_spec3D(
    spec3d_fp: PathLike, kind: Literal["geospcub"]
) -> GeoSpectrum3D: ...


def read_spec3D(spec3d_fp: PathLike, kind: Spec3DFileLiteral):
    if kind == "spcub":
        with open(spec3d_fp, "r") as f:
            json_str = f.read()
        spec_model = Spectrum3D.model_validate_json(json_str)
        return spec_model
    if kind == "geospcub":
        with open(spec3d_fp, "r") as f:
            json_str = f.read()
        spec_model = GeoSpectrum3D.model_validate_json(json_str)
        return spec_model


def read_wvl(wvl_fp: PathLike) -> WvlModel:
    """
    Read a .wvl file into a WvlModel object.

    Parameters
    ----------
    wvl_fp: File path to .wvl file.

    Returns
    -------
    WvlModel object.
    """
    file_suff = Path(wvl_fp).suffix
    if file_suff != ".wvl":
        raise FileTypeError(f"The file type should be .wvl not {file_suff}.")
    with open(wvl_fp, "r") as f:
        json_str = f.read()
    wvl_model = WvlModel.model_validate_json(json_str)
    return wvl_model



# Standard Libraries
from enum import StrEnum
from typing import Literal, Optional

# Dependencies
from pydantic import BaseModel, Field, model_validator
import rasterio as rio  # type: ignore
from rasterio.windows import Window  # type: ignore
import numpy as np

# Relative Imports
from .custom_types import PathLike
from .geospatial_models import BaseGeolocationModel
from .wvl_models import WvlModel


type Spec3DFileLiteral = Literal["spcub"] | Literal["geospcub"]


class Spec3DFileTypes(StrEnum):
    RAW = ".spcub"
    GEO = ".geospcub"


class Spectrum3D(BaseModel):
    name: str
    wavelength: WvlModel
    nrows: int
    ncols: int
    nbands: int = Field(default=0)
    raster_fp: PathLike

    @model_validator(mode="after")
    def set_nbands(self) -> "Spectrum3D":
        self.nbands = self.wavelength.nbands
        return self

    def load_raster(
        self,
        pixel_window: Optional[tuple[int, int, int, int]] = None,
        bbl: bool = True,
        nodata_value: float = -999,
    ) -> np.ndarray:
        """
        Load raster data from absolute file path stored in the file.

        Parameters
        ----------
        pixel_window: 4-tuple of int, optional
            A window tuple, (col_offset, row_offset, width, height). Default
            is None. Provide this argument to read a partial amount of the
            larger array.

        Returns
        -------
        np.ndarray
            Loaded array.
        """
        if pixel_window is not None:
            w = Window(*pixel_window)  # type: ignore
            with rio.open(self.raster_fp, "r") as f:
                arr = f.read(window=w)
        else:
            with rio.open(self.raster_fp, "r") as f:
                arr = f.read()

        arr = np.transpose(arr, (1, 2, 0))
        if bbl:
            arr = arr[:, :, self.wavelength.bbl]

        arr[arr == nodata_value] = np.nan
        return arr


class GeoSpectrum3D(Spectrum3D):
    geodata: BaseGeolocationModel



# Standard Libraries
import logging
from typing import Optional, Literal

# Dependencies
import numpy as np
import rasterio as rio  # type: ignore

# Relative Imports
from .custom_types import PathLike, Path, WvlUnit
from .wvl_models import WvlModel
from .spec1D_models import (
    Spectrum1D,
    GeoSpectrum1D,
    PointSpectrum1D,
    Spec1DFileTypes,
)
from .specgroup_models import SpectrumGroup
from .spec3D_models import Spectrum3D, GeoSpectrum3D, Spec3DFileTypes
from .geospatial_models import (
    PointModel,
    PointGeolocation,
    GeotransformModel,
    BaseGeolocationModel,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def resolve_wvlmodel(wvl: WvlModel | PathLike):
    """Returns a WvlModel from an input that is either a WvlModel or a Path."""
    # Creating Wavelength Model
    if isinstance(wvl, WvlModel):
        logger.debug("WvlModel object was provided.")
        wvlmodel = wvl
    else:
        if not Path(wvl).exists():  # Ensuring wvl file existence.
            logger.error("Wavelength file not found: %s", wvl)
            raise FileNotFoundError(f"{wvl} does not exist.")
        logger.debug("Reading wavelength file: %s", wvl)
        with open(wvl, "r") as f:
            json_str = f.read()
        wvlmodel = WvlModel.model_validate_json(json_str)

    return wvlmodel


def write_geodata(
    crs: str,
    geotransform: tuple[float, float, float, float, float, float],
    fp: PathLike,
):
    gtrans = GeotransformModel.fromgdal(geotransform)
    logger.debug("Creating geotransform model for crs=%s", crs)
    geodatamodel = BaseGeolocationModel(crs=crs, geotransform=gtrans)
    json_str = geodatamodel.model_dump_json(indent=2)
    out_path = Path(fp).with_suffix(".geodata")
    with open(out_path, "w") as f:
        f.write(json_str)
    logger.info("Wrote geodata file: %s", out_path)


def write_spec1D(
    spec_vals: np.ndarray | list,
    wvl: PathLike | WvlModel,
    name: str,
    fp: PathLike,
    *,
    location: Optional[tuple[float, float]] = None,
    location_type: Literal["pixel", "map"] = "pixel",
    geodata_fp: PathLike | None = None,
) -> None:
    """
    Writes a spectrum to one of three different file types based on provided
    spatial information.

    Parameters
    ----------
    spec_vals: ArrayLike
        Spectrum data values.
    wvl: PathLike or WvlModel
        Either a wavelength model object or a path to a .wvl file.
    name: str
        Name of the spectrum.
    fp: PathLike
        File path to save the spectrum to. File extension not included.
    location: 2-tuple of float, optional
        Location of the spectrum. Default is None.
    location_type: str, optional
        Specifies map coordinates or pixel coordinates. Either "pixel" or
        "map", respectively, default is "pixel".
    geodata_fp: PathLike, optional
        File path to saved .geodata file.

    Notes
    -----
    Depending on the input arguments, different file types will be returned:

    - If `location` is None, a .rawspec file will be created. This file has no
    spatial information associated with it, but is just a raw spectrum.

    - If `geodata_fp` is None, a .pntspec file will be created. This file
    records the pixel value that the spectrum was pulled from. In this
    scenario, `location_type` must be "pixel".

    - If `geodata_fp` is provided, a .geospec file will be created. This file
    records the pixel value and the map coordinate of the spectrum.
    """
    # Ensuring spec_vals is a list of floats for compatibility.
    if isinstance(spec_vals, np.ndarray):
        spec_vals = list(spec_vals)

    logger.debug(
        "write_spec1D called: name=%s fp=%s location=%s location_type=%s"
        "geodata_fp=%s",
        name,
        fp,
        location,
        location_type,
        geodata_fp,
    )

    wvlmodel = resolve_wvlmodel(wvl)

    # Loading geodata if it is provided
    if geodata_fp is not None:
        if not Path(geodata_fp).exists():  # Ensuring geodata file existence.
            logger.error("Geodata file not found: %s", geodata_fp)
            raise FileNotFoundError(f"{geodata_fp} does not exist.")

        logger.debug("Reading geodata file: %s", geodata_fp)
        # Creating geolocation model
        with open(geodata_fp, "r") as f:
            json_str = f.read()
        geodata = BaseGeolocationModel.model_validate_json(json_str)
    else:
        geodata = None

    if location is None:
        specmodel = Spectrum1D(
            name=name, spectrum=spec_vals, wavelength=wvlmodel
        )
        file_suffix = Spec1DFileTypes.RAW
    else:
        # Handling a georeferenced 1D spectrum.
        if geodata is not None:
            if location_type == "map":
                geopt = PointGeolocation.from_base(
                    geodata, (location[1], location[0]), location_type
                )
            elif location_type == "pixel":
                geopt = PointGeolocation.from_base(
                    geodata, location, location_type
                )

            specmodel = GeoSpectrum1D(
                name=name, spectrum=spec_vals, wavelength=wvlmodel, point=geopt
            )
            file_suffix = Spec1DFileTypes.GEO
        else:
            if location_type == "map":
                raise ValueError(
                    "If the location is in map coordinates, a file path to a "
                    ".geodata file must be provided."
                )
            specmodel = PointSpectrum1D(
                name=name,
                spectrum=spec_vals,
                wavelength=wvlmodel,
                pixel=PointModel(x=location[0], y=location[1]),
            )
            file_suffix = Spec1DFileTypes.PNT

    if Path(fp).is_dir():
        save_path = Path(fp, f"{name}").with_suffix(file_suffix)
    else:
        save_path = Path(fp).with_suffix(file_suffix)

    final_json_str = specmodel.model_dump_json(indent=2)
    with open(save_path, "w") as f:
        f.write(final_json_str)
    logger.info("Wrote spec1D file: %s", save_path)

    return None


def write_group(
    spec_group: np.ndarray | list[list[float]],
    spec_locations: list[tuple[int, int]],
    wvl: PathLike | WvlModel,
    name: str,
    fp: PathLike,
) -> None:

    wvlmodel = resolve_wvlmodel(wvl)

    spectra: list[PointSpectrum1D] = []
    if isinstance(spec_group, np.ndarray):
        # Checking size of spec_group array
        spec_data: list[list[float]] = []
        if spec_group.shape[1] != wvlmodel.nbands:
            raise ValueError(
                "The size of a spec_group array should be NxB where B is "
                f"the number of bands ({wvlmodel.nbands}). It is"
                f"currently {spec_group.shape}."
            )
        for n in range(spec_group.shape[0]):
            spec_data.append(list(spec_group[n, :]))
    else:
        spec_data = spec_group

    for n, spec in enumerate(spec_data):
        spectra.append(
            PointSpectrum1D(
                name=f"{name}_{n:04d}",
                spectrum=spec,
                wavelength=wvlmodel,
                pixel=PointModel(
                    x=spec_locations[n][0], y=spec_locations[n][1]
                ),
            )
        )

    grp_obj = SpectrumGroup(
        name=name,
        spectra=spectra,
        spectra_pts=spec_locations,
        wavelength=wvlmodel,
    )
    json_str = grp_obj.model_dump_json(indent=2)

    if Path(fp).is_dir():
        save_path = Path(fp, name).with_suffix(".specgrp")
    else:
        save_path = Path(fp).with_suffix(".specgrp")
    with open(save_path, "w") as f:
        f.write(json_str)
        logger.debug(f"Wrote SpectrumGroup to {save_path}.")

    return None


def write_spec3D(
    name: str,
    wvl: PathLike | WvlModel,
    raster_fp: PathLike,
    fp: PathLike,
    *,
    geodata_fp: Optional[PathLike] = None,
) -> None:
    wvlmodel = resolve_wvlmodel(wvl)
    with rio.open(raster_fp, "r") as f:
        height: int = f.height
        width: int = f.width
        count: int = f.count
    if count != wvlmodel.nbands:
        raise ValueError(
            f"Loaded raster data ({count}) does not have the same number of "
            f"bands as the wavelength data ({wvlmodel.nbands})."
        )
    if geodata_fp is None:
        spec3dmodel = Spectrum3D(
            name=name,
            wavelength=wvlmodel,
            nrows=height,
            ncols=width,
            raster_fp=raster_fp,
        )
        file_suffix = Spec3DFileTypes.RAW
    else:
        with open(geodata_fp, "r") as f:
            geojson_str = f.read()
        geodat = BaseGeolocationModel.model_validate_json(geojson_str)
        spec3dmodel = GeoSpectrum3D(
            name=name,
            wavelength=wvlmodel,
            nrows=height,
            ncols=width,
            raster_fp=raster_fp,
            geodata=geodat,
        )
        file_suffix = Spec3DFileTypes.GEO

    if Path(fp).is_dir():
        save_path = Path(fp, name).with_suffix(file_suffix)
    else:
        save_path = Path(fp).with_suffix(file_suffix)

    json_str = spec3dmodel.model_dump_json(indent=2)
    with open(save_path, "w") as f:
        f.write(json_str)


def write_wvl(
    wvl_values: np.ndarray | list,
    wvl_unit: WvlUnit,
    fp: PathLike,
    bbl: Optional[np.ndarray | list] = None,
) -> None:
    # Ensuring wvl_values is a list
    if isinstance(wvl_values, np.ndarray):
        wvl_values = list(wvl_values)

    # Creating WvlModel
    if bbl is None:
        bbl_vals = list(np.ones(len(wvl_values), dtype=bool))
    else:
        if isinstance(bbl, np.ndarray):
            bbl_vals = list(bbl)
        else:
            bbl_vals = bbl

    wvl = WvlModel(values=wvl_values, unit=wvl_unit, bbl=bbl_vals)

    # Dumping data to JSON
    json_string = wvl.model_dump_json(indent=2)
    out_path = Path(fp).with_suffix(".wvl")
    with open(out_path, "w") as f:
        f.write(json_string)
    logger.info("Wrote wavelength file: %s", out_path)


def write_from_object(spectrum_like_obj, save_fp: PathLike) -> None:
    json_str = spectrum_like_obj.model_dump_json(indent=2)
    suffix: str = ""
    if (
        isinstance(spectrum_like_obj, Spectrum1D)
        and not isinstance(spectrum_like_obj, PointSpectrum1D)
        and not isinstance(spectrum_like_obj, GeoSpectrum1D)
    ):
        suffix = Spec1DFileTypes.RAW
    elif isinstance(spectrum_like_obj, PointSpectrum1D):
        suffix = Spec1DFileTypes.PNT
    elif isinstance(spectrum_like_obj, GeoSpectrum1D):
        suffix = Spec1DFileTypes.GEO
    elif isinstance(spectrum_like_obj, Spectrum3D):
        suffix = Spec3DFileTypes.RAW
    elif isinstance(spectrum_like_obj, GeoSpectrum3D):
        suffix = Spec3DFileTypes.GEO
    elif isinstance(spectrum_like_obj, SpectrumGroup):
        suffix = ".specgrp"
    else:
        raise ValueError("Invalid Spectrum Like Object")

    with open(Path(save_fp).with_suffix(suffix), "w") as f:
        f.write(json_str)



"""
# `spectralio`

I/O operations for working with spectral data in python.

---

## Implemented File Types
- .rawspec
- .pntspec
- .geospec
- .specgrp
- .spcub
- .geospcub
- .wvl
"""

from .reading import (
    read_spec1D,
    read_group,
    read_spec3D,
    read_wvl,
    read_geodata,
)
from .writing import (
    write_spec1D,
    write_group,
    write_spec3D,
    write_wvl,
    write_geodata,
    write_from_object,
)
from .shapefiles import make_points, make_polygons

from .wvl_models import WvlModel
from .spec1D_models import Spectrum1D, PointSpectrum1D, GeoSpectrum1D
from .specgroup_models import SpectrumGroup
from .spec3D_models import Spectrum3D, GeoSpectrum3D
from .geospatial_models import (
    BaseGeolocationModel,
    PointGeolocation,
    RasterGeolocation,
)

__all__ = [
    "read_spec1D",
    "read_group",
    "read_spec3D",
    "read_wvl",
    "read_geodata",
    "write_spec1D",
    "write_group",
    "write_spec3D",
    "write_wvl",
    "write_geodata",
    "write_from_object",
    "WvlModel",
    "Spectrum1D",
    "PointSpectrum1D",
    "GeoSpectrum1D",
    "SpectrumGroup",
    "Spectrum3D",
    "GeoSpectrum3D",
    "make_points",
    "make_polygons",
    "BaseGeolocationModel",
    "PointGeolocation",
    "RasterGeolocation",
]



# Standard Libraries
import logging
from typing import Optional

# Dependencies
from pydantic import BaseModel, Field, model_validator
import numpy as np

# Relative Imports
from .custom_types import WvlUnit

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class WvlModel(BaseModel):
    """
    Object representing the wavelength values of a spectrum.

    Attributes
    ----------
    values: list[float]
        Wavelength values.
    unit: WvlUnit
        Unit of the wavelengths. Options are:

        - "nm": nanometers
        - "um": microns
        - "m": meters
        - "v": wavenumber

    bbl: boolean list
        Band Band List. If value in the list is False, this indicates a bad
        band. All True values indicates no bad bands.
    resolution: float
        Average spectral resolution.
    nbands: float
        Number of spectral bands.
    """

    values: list[float]
    unit: WvlUnit
    bbl: list[bool]
    resolution: float = Field(default=0.0)
    nbands: int = Field(default=0)
    ngoodbands: int = Field(default=0)

    @model_validator(mode="after")
    def set_resolution(self):
        arr = np.asarray(self.values)
        self.resolution = float((np.max(arr) - np.min(arr)) / arr.size)
        self.nbands = int(len(self.values))
        self.ngoodbands = int(np.count_nonzero(np.asarray(self.bbl)))
        return self

    @classmethod
    def default_bbl(cls, values: list[float], unit: WvlUnit) -> "WvlModel":
        return cls(values=values, unit=unit, bbl=[True] * len(values))

    @classmethod
    def fromarray(
        cls,
        values: np.ndarray,
        unit: WvlUnit,
        bbl: Optional[list[bool]] = None,
    ) -> "WvlModel":
        if bbl is None:
            bbl = [True] * values.size
        return cls(values=list(values), unit=unit, bbl=bbl)

    def __array__(self):
        return np.asarray(self.values)

    def __getitem__(self, idx):
        return np.asarray(self.values)[idx]

    def __setitem__(self, idx, value):
        arr = np.asarray(self.values)
        arr[idx] = value
        self.values = arr.tolist()

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return f"ArrayLike({self.values})"

    def asarray(self, bbl: bool = True):
        """
        Returns an array of wavelength values.

        Parameters
        ----------
        bbl: bool, optional
            If true (default), applies bad band mask to eliminate any band
            bands specified in `bbl`.
        """
        if bbl:
            return np.asarray(self.values)[np.asarray(self.bbl)]
        else:
            return np.asarray(self.values)

    def applybbl(self):
        """
        Removes bad bands from wavelength values list.
        """
        return [i for i in self.asarray(bbl=True)]

    def to_nm(self):
        """
        Converts the wavelength values to nanometers.
        """
        arr = np.asarray(self.values)
        if self.unit == "nm":
            pass
        elif self.unit == "um":
            arr *= 10**3
        elif self.unit == "m":
            arr *= 10**9
        elif self.unit == "v":
            arr = 10**9 / arr

        self.values = list(arr)
        self.unit = "nm"
        logger.debug(f"{id(self)} was converted to nm")

    def to_um(self):
        """
        Converts the wavelength values to microns.
        """
        arr = np.asarray(self.values)
        if self.unit == "nm":
            arr *= 10**-3
        elif self.unit == "um":
            pass
        elif self.unit == "m":
            arr *= 10**6
        elif self.unit == "v":
            arr = 10**6 / arr

        self.values = list(arr)
        self.unit = "um"
        logger.debug(f"{id(self)} was converted to um")

    def to_m(self):
        """
        Converts the wavelength values to meters.
        """
        arr = np.asarray(self.values)
        if self.unit == "nm":
            arr *= 10**-9
        elif self.unit == "um":
            arr *= 10**-6
        elif self.unit == "m":
            pass
        elif self.unit == "v":
            arr = 1 / arr

        self.values = list(arr)
        self.unit = "m"
        logger.debug(f"{id(self)} was converted to m")

    def to_v(self):
        """
        Converts the wavelength values to wavenumber.
        """
        arr = np.asarray(self.values)
        if self.unit == "nm":
            arr = 10**9 / arr
        elif self.unit == "um":
            arr = 10**6 / arr
        elif self.unit == "m":
            arr = 1 / arr
        elif self.unit == "v":
            pass

        self.values = list(arr)
        self.unit = "v"
        logger.debug(f"{id(self)} was converted to v")

    def convert_to(self, unit: WvlUnit):
        if unit == "nm":
            self.to_nm()
        elif unit == "um":
            self.to_um()
        elif unit == "m":
            self.to_m()
        elif unit == "v":
            self.to_v()

    def find(
        self, wvl_guess: float | int, unit: WvlUnit
    ) -> tuple[np.int32 | np.int64, float]:
        """
        Finds the wavelength and index closest to a guess value.

        Parameters
        ----------
        wvl_guess: float
            Estimate of the wavelength to find.
        unit: WvlUnit
            Unit of the wavelength guess.

        Returns
        -------
        index: int
            Index of the found wavelength.
        exact_wvl_value: float
            Exact value of the found wavelength.
        """
        _current_unit = self.unit
        self.convert_to(unit)
        arr = self.asarray()
        idx = np.argmin(abs(arr - wvl_guess))
        self.convert_to(_current_unit)
        return idx, self.values[idx]



# Standard Library
from typing import Literal, TypeAlias
from pathlib import Path
import os

WvlUnit: TypeAlias = Literal["nm", "um", "m", "v"]
type PathLike = str | Path | os.PathLike
# SpectrumLike: TypeAlias = (
#     Spectrum1D
#     | PointSpectrum1D
#     | GeoSpectrum1D
#     | Spectrum3D
#     | GeoSpectrum3D
#     | SpectrumGroup
# )



# Standard Libraries
import logging
from dataclasses import dataclass
from typing import Literal

# Dependencies
from pydantic import BaseModel, Field, model_validator

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class GeographicBoundsError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class PointModel(BaseModel):
    """Representation of a ordered pair"""

    x: float
    y: float

    def astuple(self):
        return (self.x, self.y)


@dataclass
class Bounds:
    """Representation of a bounding box"""

    left: float
    bottom: float
    right: float
    top: float


class GeotransformModel(BaseModel):
    """
    Object representing an affine geotransform matrix.

    Parameters
    ----------
    upperleft: PointModel
        Map coordinates of the upper left point of the raster.
    xres: float
        east-west resolution of the raster in map_unit/pixel.
    row_rotation: float
        Amount of rotation for the rows. Usually 0.
    yres: float
        north-south resolution of the raster in map_unit/pixel.
    col_rotation
        Amount of rotation for the columns. Usually 0.
    """

    upperleft: PointModel
    xres: float
    row_rotation: float
    yres: float
    col_rotation: float

    @classmethod
    def null(cls):
        return cls(
            upperleft=PointModel(x=0, y=0),
            xres=1,
            row_rotation=0,
            yres=-1,
            col_rotation=0,
        )

    @classmethod
    def fromgdal(
        cls, gdal_transform: tuple[float, float, float, float, float, float]
    ) -> "GeotransformModel":
        """Creates a GeotransformModel from a GDAL-style geotransform."""
        upper_left_pt = PointModel(x=gdal_transform[0], y=gdal_transform[3])
        return cls(
            upperleft=upper_left_pt,
            xres=gdal_transform[1],
            row_rotation=gdal_transform[2],
            yres=gdal_transform[5],
            col_rotation=gdal_transform[4],
        )

    def togdal(self) -> tuple[float, float, float, float, float, float]:
        """Returns a GDAL-style geotransform."""
        return (
            self.upperleft.x,
            self.xres,
            self.row_rotation,
            self.upperleft.y,
            self.yres,
            self.col_rotation,
        )

    def get_bbox(self, height: int, width: int):
        """Given the height and width of a raster, return a bounding box."""
        return Bounds(
            left=self.upperleft.x,
            bottom=self.upperleft.y + height * self.yres,
            right=self.upperleft.x + width * self.xres,
            top=self.upperleft.y,
        )

    def pixel_to_map(
        self,
        xpixel: float,
        ypixel: float,
        convention: Literal["globe", "hemi"] = "hemi",
    ) -> tuple[float, float]:
        """Convert a pixel coordinate point to a map coordinate point."""
        xmap = (
            self.upperleft.x + xpixel * self.xres + ypixel * self.row_rotation
        )
        ymap = (
            self.upperleft.y + ypixel * self.yres + xpixel * self.col_rotation
        )

        if convention == "globe":
            if xmap < 0:
                xmap += 360
        return xmap, ymap

    def map_to_pixel(self, xmap: float, ymap: float) -> tuple[float, float]:
        """Convert a map coordinate point to a pixel coordinate point."""
        _scaler = (self.xres * self.yres) - (
            self.col_rotation * self.row_rotation
        )
        xpixel = (
            (self.yres * xmap)
            - (self.upperleft.x * self.yres)
            - ((self.row_rotation * self.yres) / self.xres) * ymap
            + (self.row_rotation * self.upperleft.y)
        ) / _scaler
        ypixel = (
            (self.xres * ymap)
            - (self.upperleft.y * self.xres)
            - ((self.col_rotation * self.xres) / self.yres) * xmap
            + (self.col_rotation * self.upperleft.x)
        ) / _scaler

        if xpixel < 0:
            raise GeographicBoundsError(
                f"{xmap} is beyond the left X bound: {self.upperleft.x}"
            )
        if ypixel < 0:
            raise GeographicBoundsError(
                f"{ymap} is beyond the top bound: {self.upperleft.y}"
            )
        return xpixel, ypixel


class BaseGeolocationModel(BaseModel):
    """
    Contains all information about the geolocation of a raster or point.
    """

    crs: str
    geotransform: GeotransformModel

    @classmethod
    def null(cls):
        return cls(
            crs=(
                'GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",'
                'SPHEROID["WGS_1984",6378137.0,298.257223563]],'
                'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]]'
            ),
            geotransform=GeotransformModel.null(),
        )


class PointGeolocation(BaseGeolocationModel):
    map_point: PointModel
    pixel_point: PointModel

    @classmethod
    def from_base(
        cls,
        geoloc: BaseGeolocationModel,
        location: tuple[float, float],
        location_type: Literal["map", "pixel"],
    ) -> "PointGeolocation":
        if location_type == "map":
            map_pt = location
            pixel_pt = geoloc.geotransform.map_to_pixel(*location)
        elif location_type == "pixel":
            map_pt = geoloc.geotransform.pixel_to_map(*location)
            pixel_pt = location
        return cls(
            crs=geoloc.crs,
            geotransform=geoloc.geotransform,
            map_point=PointModel(x=map_pt[0], y=map_pt[1]),
            pixel_point=PointModel(x=pixel_pt[0], y=pixel_pt[1]),
        )


class RasterGeolocation(BaseGeolocationModel):
    height: int
    width: int
    bounds: Bounds = Field(default=Bounds(left=0, bottom=0, right=0, top=0))

    @model_validator(mode="after")
    def set_bounds(self) -> "RasterGeolocation":
        self.bounds = self.geotransform.get_bbox(self.height, self.width)
        return self
